
#include "cluster.h"
#include "curvebem.h"
#include "supermatrix.h"
#include "graphics.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <sys/times.h>
#include <time.h>

/* ************************************************************
    Demonstrate the construction of BEM-matrices on curves
   ************************************************************ */

int
main()
{
  pclustertree ct;
  pcluster root;
  pclusterfactory factory;
  pcurvebemfactory bfactory;
  psupermatrix M, Mfull;
  double (*vertex)[2];
  int *index;
  double error, norm, zeit;
  int i, n, p, nmin;
  struct tms buffer;
  clock_t t1, t2;

  n = 256;
  p = 3;

  (void) printf("==============================================\n");
  (void) printf("#     Winterschool: BEM Example              #\n");
  (void) printf("==============================================\n");
  (void) printf("# Number of basis functions?\n");
  scanf("%d", &n);
  (void) printf("# Order of the interpolation?\n");
  scanf("%d", &p);
  (void) printf("# nmin = "); scanf("%d", &nmin);

  t1 = times(&buffer);
 
  (void) printf("# Creating a curve approximation with %d intervals\n",
		n);

  vertex = (double (*)[2]) malloc((size_t) sizeof(double [2]) * n);
  assert(vertex != NULL);
  
  for(i=0; i<n; i++) {
    vertex[i][0] = sin(2.0 * M_PI * i / n);
    vertex[i][1] = cos(2.0 * M_PI * i / n);
  }

  (void) printf("# Initializing clusterfactory\n");

  factory = new_clusterfactory(n, n, 2);

  for(i=0; i<n; i++) {
    factory->x[i][0] = 0.5 * (vertex[i][0] + vertex[(i+1)%n][0]);
    factory->x[i][1] = 0.5 * (vertex[i][1] + vertex[(i+1)%n][1]);

    if(vertex[i][0] < vertex[(i+1)%n][0]) {
      factory->smin[i][0] = vertex[i][0];
      factory->smax[i][0] = vertex[(i+1)%n][0];
    }
    else {
      factory->smin[i][0] = vertex[(i+1)%n][0];
      factory->smax[i][0] = vertex[i][0];
    }

    if(vertex[i][1] < vertex[(i+1)%n][1]) {
      factory->smin[i][1] = vertex[i][1];
      factory->smax[i][1] = vertex[(i+1)%n][1];
    }
    else {
      factory->smin[i][1] = vertex[(i+1)%n][1];
      factory->smax[i][1] = vertex[i][1];
    }
  }

  (void) printf("# Creating cluster tree\n");

  index = (int *) malloc((size_t) sizeof(int) * n);
  assert(index != NULL);

  ct = create_clustertree(factory, HLIB_GEOMETRIC, nmin);
  root = ct->root;

  (void) printf("# Creating supermatrix\n");

  M = build_supermatrix_from_cluster(root, root, 2.0, 64, HLIB_MINADM);

  (void) printf("# Filling supermatrix\n");

  bfactory = new_curvebemfactory(n, ct->dof2idx, (const double (*)[2]) vertex, p, p);

  fill_curvebem_supermatrix(M, root, root, bfactory);

  t2 = times(&buffer);
  zeit = (((double)t2-t1)/CLK_TCK);
  printf("# Time for the assembly was %.2f seconds.\n",zeit);

  (void) printf("# Creating dense supermatrix\n");

  Mfull = build_full_supermatrix(root, root);

  (void) printf("# Filling dense supermatrix\n");

  fill_curvebem_supermatrix(Mfull, root, root, bfactory);

  (void) printf("# Computing approximation error\n");

  error = norm2diff_supermatrix(M, Mfull);

  (void) printf("# Approximation error: %.3g\n"
		"# Computing matrix norm\n",
		error);

  norm = norm2_supermatrix(Mfull);

  (void) printf("# Matrix norm: %.3g\n"
		"# Relative approximation error: %.3g\n",
		norm, error / norm);
  
  return 0;
}










