
#include "cluster.h"
#include "supermatrix.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

/* ************************************************************
    Demonstrate the construction of a cluster tree and supermatrix
   ************************************************************ */

int
main()
{
  pclustertree ct;
  pclusterfactory factory;
  psupermatrix M;
  double *px, *py;
  int i, n;

  n = 256;

  (void) printf("==============================================\n");
  (void) printf("#     Winterschool: Clustering Example       #\n");
  (void) printf("==============================================\n");
  (void) printf("# Number of basis functions?\n");
  scanf("%d", &n);

  (void) printf("# Creating a curve approximation with %d intervals\n", n);
  px = (double *) malloc((size_t) sizeof(double) * n);
  assert(px != NULL);

  py = (double *) malloc((size_t) sizeof(double) * n);
  assert(py != NULL);

  for(i=0; i<n; i++) {
    px[i] = sin(2.0 * M_PI * i / n);
    py[i] = cos(2.0 * M_PI * i / n);
  }

  (void) printf("# Initializing clusterfactory\n");
  factory = new_clusterfactory(n, n, 2);

  for(i=0; i<n; i++) {
    factory->x[i][0] = 0.5 * (px[i] + px[(i+1)%n]);
    factory->x[i][1] = 0.5 * (py[i] + py[(i+1)%n]);

    if(px[i] < px[(i+1)%n]) {
      factory->smin[i][0] = px[i];
      factory->smax[i][0] = px[(i+1)%n];
    }
    else {
      factory->smin[i][0] = px[(i+1)%n];
      factory->smax[i][0] = px[i];
    }

    if(py[i] < py[(i+1)%n]) {
      factory->smin[i][1] = py[i];
      factory->smax[i][1] = py[(i+1)%n];
    }
    else {
      factory->smin[i][1] = py[(i+1)%n];
      factory->smax[i][1] = py[i];
    }
  }

  (void) printf("# Creating cluster tree\n");
  ct = create_clustertree(factory, HLIB_GEOMETRIC, 1);

  if(n <= 16) {
    (void) printf("\nCluster tree:\n");
    print_cluster(ct->root, ct->dof2idx, 0);
  }

  (void) printf("# Creating supermatrix\n");
  M = build_supermatrix_from_cluster(ct->root, ct->root, 2.0, 4, HLIB_MINADM);

  (void) printf("# Printing the matrix structure to 'matrix.ps'\n");
  outputrank_supermatrix(M, "matrix.ps", 0, 0);

  return 0;
}
