
#include "cluster.h"
#include "blockcluster.h"
#include "fem2d.h"
#include "sparsematrix.h"
#include "supermatrix.h"
#include "krylov.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <sys/times.h>
#include <time.h>

/* ************************************************************
    Demonstrate the construction of FEM matrices
   ************************************************************ */

int
main()
{
  pclustertree ct;
  pcluster root;
  pblockcluster bc;
  pgrid2d g2;
  psparsematrix S;
  psupermatrix M;
  double norm,zeit;
  int nx, ny, k;
  struct tms buffer;
  clock_t t1, t2;
  char buf[80];

  nx = 32;
  ny = 32;
  k = 4;

  (void) printf("#==============================================#\n");
  (void) printf("#     FEM Example                              #\n");
  (void) printf("#==============================================#\n");
  (void) printf("# Number of unknowns in x direction = "); 
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%d", &nx);
  (void) printf("# Number of unknowns in y direction = "); 
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%d", &ny);
  (void) printf("# Blockwise rank for the H-matrix k = "); 
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%d", &k);

  (void) printf("# Creating triangulation of the unit square\n# using a %d x %d grid\n", nx+2, ny+2);
  g2 = unitsquare_grid2d(nx+1, ny+1);
  
  (void) printf("# Creating cluster tree\n");
  ct = buildcluster_grid2d(g2, HLIB_GEOMETRIC, 32);
  root = ct->root;
  
  (void) printf("# Creating blockcluster tree\n");
  bc = build_blockcluster(root, root, 2.0, 0, 0, 50, 1);
  
  (void) printf("# Building the sparse matrix\n");
  S = laplace_grid2d(g2, ct);
  
  (void) printf("# Creating %d x %d supermatrix structure\n",
		root->size, root->size);
  M = buildsupermatrixfrom_blockcluster(bc, k, 0.0);
  
  (void) printf("# Converting sparse matrix to supermatrix\n");
  convertsparse2_supermatrix(S, M);
  
  (void) printf("# Printing the matrix structure to 'matrix.ps'\n");
  outputrank_supermatrix(M, "matrix.ps", 0, 0);
  
  (void) printf("# Computing Cholesky decomposition\n");
  t1 = times(&buffer);
  choleskydecomposition_supermatrix(M);
  t2 = times(&buffer);
  zeit = (((double)t2-t1)/CLK_TCK);
  (void) printf("# Time for the decomposition was %.2f seconds.\n",zeit);

  (void) printf("# Computing inversion error\n");
  norm = norm2cholesky_sparsematrix(M, S);

  (void) printf("# Result: |(LL^{T})^{-1} * M - I|_2 = %.3e\n", norm);
  (void) printf("#==============================================#\n");
  
  return 0;
}
