
/* ------------------------------------------------------------
   This is the file "h2conversion.h" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef H2Conversion
#define H2Conversion

#include "basic.h"
#include "supermatrix.h"

/* Find an orthogonal row cluster basis that approximates a supermatrix.
   s: Matrix
   root: Root of the corresponding row cluster tree
   eps: Desired precision
   kmax: Maximal rank of the cluster basis
   strategy: Strategy for truncating singular values */
pclusterbasis
buildrow2_supermatrix(psupermatrix s, pccluster root,
		      double eps, int kmax, TruncationStrategy strategy);

/* Find an orthogonal column cluster basis that approximates a supermatrix.
   s: Matrix
   root: Root of the corresponding column cluster tree
   eps: Desired precision
   kmax: Maximal rank of the cluster basis
   strategy: Strategy for truncating singular values */
pclusterbasis
buildcol2_supermatrix(psupermatrix s, pccluster ct,
		      double eps, int kmax, TruncationStrategy strategy);

/* Create a supermatrix containing only fullmatrix and uniformmatrix blocks
   (i.e., an H^2 matrix) by orthogonal projection of a general supermatrix.
   s: Source supermatrix
   mixed: Allow mixing of rkmatrix and uniformmatrix blocks to improve performance
   row: Row cluster basis, has to be orthogonal
   rproj: Row cluster projection (optional)
   col: Column cluster basis, has to be orthogonal
   cproj: Column cluster basis projection (optional) */
psupermatrix
project2bases_supermatrix(psupermatrix s, int mixed,
			  pclusterbasis row, pclusteroperator rproj,
			  pclusterbasis col, pclusteroperator cproj);

/* Approximate a supermatrix by an H^2 matrix.
   s: Source supermatrix
   row: Root of the corresponding row cluster tree
   col: Root of the corresponding column cluster tree
   eps: Desired precision in the Frobenius norm
   kmax: Maximal rank of the cluster bases to be created
   strategy: Strategy for truncating singular values */
psupermatrix
buildh2_supermatrix(psupermatrix s, pccluster row, pccluster col,
		    double eps, int kmax, TruncationStrategy strategy);

/* Approximate a supermatrix by a symmetric H^2 matrix. Obviously, this
   can be expected to yield reasonable results if the original matrix
   is symmetric.
   s: Source supermatrix
   row: Root of the corresponding row cluster tree
   col: Root of the corresponding column cluster tree
   eps: Desired precision in the Frobenius norm
   kmax: Maximal rank of the cluster bases to be created
   strategy: Strategy for truncating singular values */
psupermatrix
buildh2symm_supermatrix(psupermatrix s, pccluster ct,
			double eps, int kmax, TruncationStrategy strategy);

/* Create an H^2 matrix with orthogonal cluster bases
   s: Source supermatrix
   eps: Desired precision in the Frobenius norm
   strategy: Strategy for truncating singular values */
psupermatrix
orthogonalize_supermatrix(psupermatrix s,
			  double eps, TruncationStrategy strategy);

#endif
