
/* ------------------------------------------------------------
   This is the file "hcoarsening.c" of the HLib package.

   Copyright 1999 Lars Grasedyck and Steffen Boerm
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "lapack.h"
#include "hcoarsening.h"
#include <assert.h>
#include <math.h>

void
compress_supermatrix(psupermatrix s, double eps){
  int i,j;
  prkmatrix r;
  pfullmatrix f;
  long int sizerk=0,
    sizefull = (long int)s->rows * (long int)s->cols * sizeof(double);
 
  if(s->s != 0x0){
    for(i=0; i<s->block_rows*s->block_cols; i++){
      compress_supermatrix(s->s[i],eps);
    }
  }
  
  if(s->s!=0x0){
    for(j=0,i=0; i<s->block_rows*s->block_cols; i++)
      if(s->s[i]->s != 0x0) j=1;
    if(j==0){
      sizerk = getsize_supermatrix(s);
      if(sizerk > sizefull){
	f = new_fullmatrix(s->rows,s->cols);
	convertsuper2_fullmatrix(f,s);
	for(i=0; i<s->block_rows*s->block_cols; i++)
	  del_supermatrix(s->s[i]);
	free(s->s);
	s->s = 0x0;
	s->f = f;
	s->r = 0x0;
      }
    }
  }
  
  if(s->r!=0x0){
    r = new_rkmatrix(0,s->rows,s->cols);
    r->local_eps = eps;
    rk_arithmetics_set_adaptive(1);
    addpart2_rkmatrix(r,0,0,s->r,0,0,r->rows,r->cols);
    rk_arithmetics_set_adaptive(0);
    del_rkmatrix(s->r);
    s->r = r;
  }
}

long int 
counttrunc_supermatrix(psupermatrix s){
  long int c = 0;
  int i;
  if(s->s != 0x0){
    for(i=0; i<s->block_rows*s->block_cols; i++){
      c += counttrunc_supermatrix(s->s[i]);
    }
  }
  if(s->f != 0x0) c += s->rows*s->cols;
  if(s->r != 0x0) c += s->r->kt*s->r->kt*(s->rows+s->cols);
  return c;
}

void
onthefly_aca_coarsen_supermatrix(psupermatrix s, pccluster row, pccluster col,
				 psurfacebemfactory bfactory,
				 double epsaca, int kmax,
				 double **buf, int *bufsize, 
				 double eps, int diag){
  int i,j,ad;
  prkmatrix r;
  pfullmatrix f;
  long int sizerk=0,sizesuper=0,
    sizefull = (long int)s->rows * (long int)s->cols * sizeof(double);
  ad = rk_arithmetics_adaptive();
  assert(s->u == NULL);

  if(s->s != 0x0){
    for(i=0; i<s->block_rows; i++){
      for(j=0; j<s->block_cols; j++){
	if(i==j){
	  onthefly_aca_coarsen_supermatrix(s->s[i+j*s->block_rows],
					   row->son[i],col->son[j],
					   bfactory,epsaca,kmax,buf,
					   bufsize,eps,diag);
	}else{
	  onthefly_aca_coarsen_supermatrix(s->s[i+j*s->block_rows],
					   row->son[i],col->son[j],
					   bfactory,epsaca,kmax,buf,
					   bufsize,eps,0);
	}
      }
    }
    sizesuper = getsize_supermatrix(s);
    for(j=0,i=0; i<s->block_rows*s->block_cols; i++)
      if(s->s[i]->s != 0x0 || diag==1) j=1;
    if(j==0){
      /* try to coarsen */
      r = new_rkmatrix(0,s->rows,s->cols);
      r->local_eps = eps;
      rk_arithmetics_set_adaptive(1);
      convertsuper2_rkmatrix(s, r);
      rk_arithmetics_set_adaptive(0);
      sizerk = getsize_rkmatrix(r);
      if(sizerk < sizesuper){
	s->r = r;
	for(i=0; i<s->block_rows*s->block_cols; i++) 
	  del_supermatrix(s->s[i]);
	free(s->s);
	s->s = 0x0;
      }else{
	del_rkmatrix(r);
      }
    }
    /* try to replace by fullmatrix */
    sizesuper = getsize_supermatrix(s);
    if(sizesuper > sizefull){
      f = new_fullmatrix(s->rows,s->cols);
      convertsuper2_fullmatrix(f,s);
      if(s->s != 0x0){
	for(i=0; i<s->block_rows*s->block_cols; i++)
	  del_supermatrix(s->s[i]);
	free(s->s);
      }else{
	del_rkmatrix(s->r);
      }
      s->s = 0x0;
      s->f = f;
      s->r = 0x0;
    }
  }else{
    if(s->r != 0x0){
      assert(s->r->rows == row->size);
      assert(s->r->cols == col->size);
      
      bufaca2_rkmatrix(bfactory->nearfield_entry, bfactory,
		       epsaca, kmax, row->start, col->start, buf, bufsize,
		       s->r, HLIB_ACA_DEFAULT);
      
      r = new_rkmatrix(0,s->rows,s->cols);
      r->local_eps = eps;
      rk_arithmetics_set_adaptive(1);
      addpart2_rkmatrix(r,0,0,s->r,0,0,r->rows,r->cols);
      rk_arithmetics_set_adaptive(ad);
      del_rkmatrix(s->r);
      s->r = r;
    }else{     
      assert(s->f != NULL);
      bfactory->integrate_nearfield(row, col, bfactory, s->f->e, s->f->rows);
    }
  }  
}

psupermatrix
coarsenh2_supermatrix(psupermatrix s, double eps, int diag){
  int i,j,ad;
  prkmatrix r;
  pfullmatrix f;
  psupermatrix sc = 0x0;
  long int sizerk=0,sizesuper=0,
    sizefull = (long int)s->rows * (long int)s->cols * sizeof(double);
  ad = rk_arithmetics_adaptive();
  
  if(s->s != 0x0){
    sc = new_supermatrix(s->block_rows,s->block_cols,s->rows,s->cols,
			 (puniformmatrix)0, (prkmatrix)0, (pfullmatrix)0);
    for(i=0; i<s->block_rows; i++){
      for(j=0; j<s->block_cols; j++){
	if(i==j){
	  sc->s[i+j*s->block_rows] = 
	    coarsenh2_supermatrix(s->s[i+j*s->block_rows],eps,diag);
	}else{
	  sc->s[i+j*s->block_rows] = 
	    coarsenh2_supermatrix(s->s[i+j*s->block_rows],eps,0);
	}
      }
    }
    sizesuper = getsize_supermatrix(sc);
    for(j=0,i=0; i<sc->block_rows*sc->block_cols; i++)
      if(sc->s[i]->s != 0x0 || diag==1) j=1;
    if(j==0){
      /* try to coarsen */
      r = new_rkmatrix(0,sc->rows,sc->cols);
      r->local_eps = eps;
      rk_arithmetics_set_adaptive(1);
      convertsuper2_rkmatrix(sc, r);
      rk_arithmetics_set_adaptive(0);
      sizerk = getsize_rkmatrix(r);
      if(sizerk < sizesuper){
	sc->r = r;
	for(i=0; i<sc->block_rows*sc->block_cols; i++) 
	  del_supermatrix(sc->s[i]);
	free(sc->s);
	sc->s = 0x0;
      }else{
	del_rkmatrix(r);
      }
    }
    /* try to replace by fullmatrix */
    if(diag==1){
      sizesuper = getsize_supermatrix(sc);
      if(sizesuper > sizefull){
	f = new_fullmatrix(sc->rows,sc->cols);
	convertsuper2_fullmatrix(f,sc);
	if(sc->s != 0x0){
	  for(i=0; i<sc->block_rows*sc->block_cols; i++)
	    del_supermatrix(sc->s[i]);
	  free(sc->s);
	}else{
	  if(sc->r != 0x0) del_rkmatrix(sc->r);
	  if(sc->u != 0x0) del_uniformmatrix(sc->u);
	}
	sc->s = 0x0;
	sc->f = f;
	sc->r = 0x0;
	sc->u = 0x0;
      }
    }
  }else{
    if(s->u != 0x0){
      r = new_rkmatrix(0,s->rows,s->cols);
      r->local_eps = eps;
      approxuniform2_rkmatrix(s->u, r, eps, 100, HLIB_EUCLIDEAN_RELATIVE);

      sc = new_supermatrix(1,1,s->rows,s->cols,(puniformmatrix)0, r, 
			   (pfullmatrix)0);
    }
    if(s->r != 0x0){
      r = new_rkmatrix(0,s->rows,s->cols);
      r->local_eps = eps;
      rk_arithmetics_set_adaptive(1);
      addpart2_rkmatrix(r,0,0,s->r,0,0,r->rows,r->cols);
      rk_arithmetics_set_adaptive(ad);
      sc = new_supermatrix(1,1,s->rows,s->cols,(puniformmatrix)0, r, (pfullmatrix)0);
    }
    if(s->f != 0x0){
      if(diag==0){
	r = new_rkmatrix(0,s->rows,s->cols);
	r->local_eps = eps;
	rk_arithmetics_set_adaptive(1);
	convertsuper2_rkmatrix(s, r);
	rk_arithmetics_set_adaptive(ad);
	sc = new_supermatrix(1,1,s->rows,s->cols,
			     (puniformmatrix)0, r, (pfullmatrix)0);
      }else{
	f = copy_fullmatrix(s->f);
	sc = new_supermatrix(1,1,s->rows,s->cols,
			     (puniformmatrix)0, (prkmatrix)0, f);
      }
    }
  }
  return sc;
}

void
coarsen_supermatrix(psupermatrix s, double eps, int diag){
  int i,j,ad;
  prkmatrix r;
  pfullmatrix f;
  long int sizerk=0,sizesuper=0,
    sizefull = (long int)s->rows * (long int)s->cols * sizeof(double);
  ad = rk_arithmetics_adaptive();
  assert(s->u == 0x0);

  if(s->s != 0x0){
    for(i=0; i<s->block_rows; i++){
      for(j=0; j<s->block_cols; j++){
	if(i==j){
	  coarsen_supermatrix(s->s[i+j*s->block_rows],eps,diag);
	}else{
	  coarsen_supermatrix(s->s[i+j*s->block_rows],eps,0);
	}
      }
    }
    sizesuper = getsize_supermatrix(s);
    for(j=0,i=0; i<s->block_rows*s->block_cols; i++)
      if(s->s[i]->s != 0x0 || diag==1) j=1;
    if(j==0){
      /* try to coarsen */
      r = new_rkmatrix(0,s->rows,s->cols);
      r->local_eps = eps;
      rk_arithmetics_set_adaptive(1);
      convertsuper2_rkmatrix(s, r);
      rk_arithmetics_set_adaptive(ad);
      sizerk = getsize_rkmatrix(r);
      if(sizerk < sizesuper){
	s->r = r;
	for(i=0; i<s->block_rows*s->block_cols; i++) 
	  del_supermatrix(s->s[i]);
	free(s->s);
	s->s = 0x0;
      }else{
	del_rkmatrix(r);
      }
    }
    /* try to replace by fullmatrix */
    sizesuper = getsize_supermatrix(s);
    if(sizesuper > sizefull){
      f = new_fullmatrix(s->rows,s->cols);
      convertsuper2_fullmatrix(f,s);
      if(s->s != 0x0){
	for(i=0; i<s->block_rows*s->block_cols; i++)
	  del_supermatrix(s->s[i]);
	free(s->s);
      }else{
	del_rkmatrix(s->r);
      }
      s->s = 0x0;
      s->f = f;
      s->r = 0x0;
    }
  }else{
    if(s->r != 0x0){
      r = new_rkmatrix(0,s->rows,s->cols);
      r->local_eps = eps;
      rk_arithmetics_set_adaptive(1);
      addpart2_rkmatrix(r,0,0,s->r,0,0,r->rows,r->cols);
      rk_arithmetics_set_adaptive(ad);
      del_rkmatrix(s->r);
      s->r = r;
    }
  }
}

void
coarsen2_supermatrix(psupermatrix s, double eps, int diag, int nmax){
  int i,j,ad;
  prkmatrix r;
  pfullmatrix f;
  long int sizerk=0,sizesuper=0,
    sizefull = (long int)s->rows * (long int)s->cols * sizeof(double);
  ad = rk_arithmetics_adaptive();
  assert(s->u == 0x0);

  if(s->s != 0x0){
    for(i=0; i<s->block_rows; i++){
      for(j=0; j<s->block_cols; j++){
	if(i==j){
	  coarsen2_supermatrix(s->s[i+j*s->block_rows],eps,diag,nmax);
	}else{
	  coarsen2_supermatrix(s->s[i+j*s->block_rows],eps,0,nmax);
	}
      }
    }
    sizesuper = getsize_supermatrix(s);
    for(j=0,i=0; i<s->block_rows*s->block_cols; i++)
      if(s->s[i]->s != 0x0 || diag==1) j=1;
    if(j==0 && s->rows<nmax && s->cols<nmax){
      /* try to coarsen */
      r = new_rkmatrix(0,s->rows,s->cols);
      r->local_eps = eps;
      rk_arithmetics_set_adaptive(1);
      convertsuper2_rkmatrix(s, r);
      rk_arithmetics_set_adaptive(ad);
      sizerk = getsize_rkmatrix(r);
      if(sizerk < sizesuper){
	s->r = r;
	for(i=0; i<s->block_rows*s->block_cols; i++) 
	  del_supermatrix(s->s[i]);
	free(s->s);
	s->s = 0x0;
      }else{
	del_rkmatrix(r);
      }
      /* try to replace by fullmatrix */
      sizesuper = getsize_supermatrix(s);
      if(sizesuper > sizefull){
	f = new_fullmatrix(s->rows,s->cols);
	convertsuper2_fullmatrix(f,s);
	if(s->s != 0x0){
	  for(i=0; i<s->block_rows*s->block_cols; i++)
	    del_supermatrix(s->s[i]);
	  free(s->s);
	}else{
	  del_rkmatrix(s->r);
	}
	s->s = 0x0;
	s->f = f;
	s->r = 0x0;
      }
    }
  }else{
    if(s->r != 0x0){
      r = new_rkmatrix(0,s->rows,s->cols);
      r->local_eps = eps;
      rk_arithmetics_set_adaptive(1);
      addpart2_rkmatrix(r,0,0,s->r,0,0,r->rows,r->cols);
      rk_arithmetics_set_adaptive(ad);
      del_rkmatrix(s->r);
      s->r = r;
    }
  }
}

psupermatrix
coarsecopy_supermatrix(psupermatrix s, int k, int minsize){
  int i,j;
  int block_rows,block_cols,rows,cols;
  psupermatrix sc;
  psupermatrix s_el;
  prkmatrix r,r2;
  puniformmatrix u;
  pfullmatrix f;
  assert(s != 0x0);

  rows = s->rows;
  cols = s->cols;
  block_rows = s->block_rows;
  block_cols = s->block_cols;

  if(s->s != 0x0){
    if(rows <= minsize || cols <= minsize){
      f = new_fullmatrix(rows,cols);
      r = 0x0;
      u = 0x0;
      convertsuper2_fullmatrix(f,s);
      sc = new_supermatrix(block_rows,block_cols,rows,cols,u,r,f);
    }else{
      r = 0x0;
      f = 0x0;
      u = 0x0;
      sc = new_supermatrix(block_rows,block_cols,rows,cols,u,r,f);
      for(i=0; i<block_rows; i++){
	for(j=0; j<block_cols; j++){
	  s_el = s->s[i+j*block_rows];
	  sc->s[i+j*block_rows] = coarsecopy_supermatrix(s_el,k,minsize);
	}
      }
    }
  }else{
    if(s->r != 0x0){
      f = 0x0;
      u = 0x0;
      r = new_rkmatrix(k,rows,cols);
      convertrk2_rkmatrix(s->r,r,0,0);
    }else{
      if(s->f != 0x0){
	u = 0x0;
	r = 0x0;
	f = copy_fullmatrix(s->f);
      }else{
	u = 0x0;
	r = new_rkmatrix(k,rows,cols);
	r2 = new_rkmatrix(s->u->row->kt,rows,cols);
	convertuniform2_rkmatrix(s->u, r2);
	convertrk2_rkmatrix(r2,r,0,0);
	del_rkmatrix(r2);
	f = 0x0;
      }
    }
    sc = new_supermatrix(1,1,rows,cols,u,r,f);
  }
  sc->l = s->l;
  return sc;
}

psupermatrix
cut_supermatrix(psupermatrix s, int k){
  int i,j;
  int block_rows,block_cols,rows,cols;
  psupermatrix sc;
  psupermatrix s_el;
  pfullmatrix f=0x0;
  prkmatrix r=0x0;
  assert(s != 0x0);
  assert(s->u == 0x0);

  rows = s->rows;
  cols = s->cols;
  block_rows = s->block_rows;
  block_cols = s->block_cols;

  if(s->s != 0x0){
    sc = new_supermatrix(block_rows,block_cols,rows,cols,0x0,0x0,0x0);
    for(i=0; i<block_rows; i++){
      for(j=0; j<block_cols; j++){
	s_el = s->s[i+j*block_rows];
	sc->s[i+j*block_rows] = cut_supermatrix(s_el,k);
      }
    }
  }else{
    if(s->r != 0x0){
      r = new_rkmatrix(k,rows,cols);
      r->kt = k;
      i = k*r->cols; 
      if(s->r->kt<k){
	i = s->r->kt*r->cols;
	r->kt = s->r->kt;
      }
      if(i>0) dcopy_(&i,s->r->b,eins_,r->b,eins_);
      i = k*r->rows; 
      if(s->r->kt<k){
	i = s->r->kt*r->rows;
      }
      if(i>0) dcopy_(&i,s->r->a,eins_,r->a,eins_);
    }else{
      if(s->f != 0x0){
	f = copy_fullmatrix(s->f);
      }
    }
    sc = new_supermatrix(1,1,rows,cols,0,r,f);
  }
  sc->l = s->l;
  return sc;
}


