
/* ------------------------------------------------------------
   This is the file "hcoarsening.h" of the HLib package.

   Copyright 1999 Lars Grasedyck and Steffen Boerm
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef HCoarsening
#define HCoarsening

#include "basic.h"
#include "fullmatrix.h"
#include "rkmatrix.h"
#include "uniformmatrix.h"
#include "sparsematrix.h"
#include "surfacebem.h"

#include <stdio.h>
#include <stdlib.h>

/* create a new supermatrix that has the same structure 
   and content as s, change the minimal blocksize to minsize, 
   and truncate all rkmatrix subblocks to rank at most k */
psupermatrix
coarsecopy_supermatrix(psupermatrix s, int k, int minsize);
psupermatrix
cut_supermatrix(psupermatrix s, int k);

void
compress_supermatrix(psupermatrix s, double eps);
void
coarsen_supermatrix(psupermatrix s, double eps,int diag);
void
coarsen2_supermatrix(psupermatrix s, double eps,int diag, int nmax);

/* compute a coarse H-matrix approximation to an H or H2-matrix
   with relative error eps, diag=1 if diagonal blocks
   are assumed to be regular */
psupermatrix
coarsenh2_supermatrix(psupermatrix s, double eps, int diag);

/* compute a coarse H-matrix approximation to an H-matrix s
   implicitly built by aca with blockwise relative euclidean
   error epsaca and coarsened to relative error eps, i.e., 
   assemble the stiffness matrix and compress it on-the-fly.
   kmax = maximal rank (aca),
   buf/bufsize = buffer for aca, e.g., buf[0]=NULL, bufsize=0,
   bfactory = bem factory for aca,
   diag = 1 for matrices regular on the diagonal,
   row/col = root cluster.
   Call this procedure with an H-matrix s whose blocks are
   not filled (k=0).
*/
void
onthefly_aca_coarsen_supermatrix(psupermatrix s, pccluster row, pccluster col,
				 psurfacebemfactory bfactory,
				 double epsaca, int kmax,
				 double **buf, int *bufsize, 
				 double eps, int diag);

#endif
