
/* ------------------------------------------------------------
   This is the file "ie1d.c" of the HLib package.

   Copyright 1999 Lars Grasedyck and Steffen Boerm
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"
#include "ie1d.h"

double  
get_entry(int n, int i, int j){
  double h;
  double value;
  double dist;

  h = 1.0/((double)n);
  dist = h*(fabs((double)(i-j))-1.0);

  if(i==j){
    value = -1.5*h*h + h*h*log(h);
  }else{
    value = -1.5*h*h + 0.5*(dist+2.0*h)*(dist+2.0*h)*log(dist+2.0*h)
      - (dist+1.0*h)*(dist+1.0*h)*log(dist+1.0*h);
    if(fabs(dist)>1e-8) value += 0.5*(dist)*(dist)*log(dist);
  }
  return value;
}

pfullmatrix 
fill_fullmatrix(int n, int size_n, int size_m, 
		int startindex_n, int startindex_m){
  int i,j;
  pfullmatrix f;
  f = new_fullmatrix(size_n,size_m);
  for(i=0; i<size_n; i++)
    for(j=0; j<size_m; j++)
      f->e[i+j*size_n] = get_entry(n,startindex_n+i,startindex_m+j);
  return f;
}

pfullmatrix 
fillsym_fullmatrix(int n, int size_n, int size_m, 
		   int startindex_n, int startindex_m){
  int i,j;
  pfullmatrix f;
  f = new_fullmatrix(size_n,size_m);
  for(i=0; i<size_n; i++)
    for(j=0; j<=i; j++)
      f->e[i+j*size_n] = get_entry(n,startindex_n+i,startindex_m+j);
  return f;
}

double  
get_taylorentry_a(int n, int i, int nu, double x0){
  double a,b;
  double value;
  double tmp;
  int j;

  a = ((double)i)/((double) n);
  b = ((double)i+1.0)/((double) n);

  tmp = b-x0;
  for(j=0; j<nu; j++) tmp *= (b-x0);
  value = tmp / ((double)nu +1.0); 
  tmp = a-x0;
  for(j=0; j<nu; j++) tmp *= (a-x0);
  value -= tmp / ((double)nu +1.0);

  return value;
}

double  
get_taylorentry_b(int n, int i, int nu, double x0){
  double a,b;
  double value;
  double tmp;
  int j;

  a = ((double)i)/((double) n);
  b = ((double)i+1.0)/((double) n);

  if(nu==0){
    value = a-b + (b-x0)*log(fabs(b-x0)) - (a-x0)*log(fabs(a-x0));
  }else{
    if(nu==1){
      value = log(fabs(x0-a)) - log(fabs(x0-b));
    }else{
      tmp = x0-b;
      for(j=0; j<nu-2; j++) tmp *= (x0-b);
      value = 1.0/tmp;
      tmp = x0-a;
      for(j=0; j<nu-2; j++) tmp *= (x0-a);
      value -= 1.0/tmp;
      value = value / ((double) nu*(nu-1));
      if(nu%2==0) value *= -1.0;
    }
  }

  return value;
}

prkmatrix
fill_rkmatrix(int n, int k, int size_n, int size_m, 
	      int startindex_n, int startindex_m){
  int i,kt;
  double x0;
  prkmatrix r;
  
  r = new_rkmatrix(k,size_n,size_m);
  r->kt = k;
  x0 = ((double)(2*startindex_n+size_n))/((double)2.0*n);
  for(i=0; i<size_n; i++)
    for(kt=0; kt<k; kt++)
      r->a[i+kt*size_n] = get_taylorentry_a(n,i+startindex_n,kt,x0);
  for(i=0; i<size_m; i++)
    for(kt=0; kt<k; kt++)
      r->b[i+kt*size_m] = get_taylorentry_b(n,i+startindex_m,kt,x0);

  return r;
}

int
check_admissible(int n0, int n, int m0, int m){
  int dist,dist2;
  int diam;
  dist = m0 - n0 - n;
  dist2 = n0 - m0 - m;
  if(dist2>dist) dist = dist2;
  diam = n; 
  if(m>n) diam = m;
  if(diam <= dist) return 1;
  return 0;
}

psupermatrix 
fillpart_supermatrix(int n, int k, int size_n, int size_m, 
		     int startindex_n, int startindex_m, int nmin){
  int nhalf,mhalf;
  psupermatrix s;
  pfullmatrix f;
  prkmatrix r;

  if(check_admissible(startindex_n,size_n,startindex_m,size_m)==1){
    /* admissible */
    r = fill_rkmatrix(n,k,size_n,size_m,startindex_n,startindex_m);
    s = new_supermatrix(1,1,size_n,size_m,0x0,r,0x0);
  }else{
    if(size_n<=nmin || size_m<=nmin){
      /* inadmissible and small */
      f = fill_fullmatrix(n,size_n,size_m,startindex_n,startindex_m);
      s = new_supermatrix(1,1,size_n,size_m,0x0,0x0,f);
    }else{
      /* inadmissible and large  */
      s = new_supermatrix(2,2,size_n,size_m,0x0,0x0,0x0);
      nhalf = size_n/2;
      mhalf = size_m/2;
      s->s[0] = fillpart_supermatrix(n,k,nhalf       ,mhalf       ,
				     startindex_n      ,startindex_m, nmin);
      s->s[1] = fillpart_supermatrix(n,k,size_n-nhalf,mhalf       ,
				     startindex_n+nhalf,startindex_m, nmin);
      s->s[2] = fillpart_supermatrix(n,k,nhalf       ,size_m-mhalf,
				     startindex_n      ,startindex_m+mhalf, nmin);
      s->s[3] = fillpart_supermatrix(n,k,size_n-nhalf,size_m-mhalf,
				     startindex_n+nhalf,startindex_m+mhalf, nmin);
    }  
  }
  return s;
}

psupermatrix 
fill_supermatrix(int n, int k, int nmin){
  psupermatrix s;
  s = fillpart_supermatrix(n,k,n,n,0,0,nmin);
  return s;
}

psupermatrix 
fillsympart_supermatrix(int n, int k, int size_n, int size_m, 
			int startindex_n, int startindex_m, int nmin){
  int nhalf,mhalf;
  psupermatrix s;
  pfullmatrix f;
  prkmatrix r;
  
  if(check_admissible(startindex_n,size_n,startindex_m,size_m)==1
     || startindex_n < startindex_m){
    /* admissible */
    if(startindex_n < startindex_m){
      r = new_rkmatrix(k,size_n,size_m);
    }else{
      r = fill_rkmatrix(n,k,size_n,size_m,startindex_n,startindex_m);
    }
    s = new_supermatrix(1,1,size_n,size_m,0x0,r,0x0);
  }else{
    if(size_n<=nmin || size_m<=nmin){
      /* inadmissible and small */
      if(startindex_n != startindex_m){
	f = fill_fullmatrix(n,size_n,size_m,startindex_n,startindex_m);
      }else{
	f = fillsym_fullmatrix(n,size_n,size_m,startindex_n,startindex_m);
      }
      s = new_supermatrix(1,1,size_n,size_m,0x0,0x0,f);
    }else{
      /* inadmissible and large  */
      s = new_supermatrix(2,2,size_n,size_m,0x0,0x0,0x0);
      nhalf = size_n/2;
      mhalf = size_m/2;
      s->s[0] = fillsympart_supermatrix(n,k,nhalf       ,mhalf       ,
				     startindex_n      ,startindex_m, nmin);
      s->s[1] = fillsympart_supermatrix(n,k,size_n-nhalf,mhalf       ,
				     startindex_n+nhalf,startindex_m, nmin);
      s->s[2] = fillsympart_supermatrix(n,k,nhalf       ,size_m-mhalf,
				     startindex_n      ,startindex_m+mhalf, nmin);
      s->s[3] = fillsympart_supermatrix(n,k,size_n-nhalf,size_m-mhalf,
				     startindex_n+nhalf,startindex_m+mhalf, nmin);
    }  
  }
  return s;
}

psupermatrix 
fillsym_supermatrix(int n, int k, int nmin){
  psupermatrix s;
  s = fillsympart_supermatrix(n,k,n,n,0,0,nmin);
  return s;
}

double  
get_rhsentry(int n, int i, int nr){
  double a,b;
  double value=0.0;
  a = ((double)i)/((double) n);
  b = ((double)i+1.0)/((double) n);
  if(nr==1){
    value = -1.5*(b - a);
    if(fabs(b)>1e-8) value += 0.5*b*b*log(b);
    if(fabs(a)>1e-8) value -= 0.5*a*a*log(a);
    if(fabs(1.0-b)>1e-8) value -= 0.5*(1.0-b)*(1.0-b)*log(1.0-b);
    if(fabs(1.0-a)>1e-8) value += 0.5*(1.0-a)*(1.0-a)*log(1.0-a); 
  }
  if(nr==2){
    value = -0.25*(b - a + b*b - a*a);
    if(fabs(b)>1e-8) value += (b*b*b*log(b))/6.0;
    if(fabs(a)>1e-8) value -= (a*a*a*log(a))/6.0;
    value -= (b*b*b - a*a*a)/18.0;
    if(fabs(1.0-b)>1e-8) value += (1.0-b)*(1.0-b)*(1.0-b)*log(1.0-b)/6.0;
    if(fabs(1.0-a)>1e-8) value -= (1.0-a)*(1.0-a)*(1.0-a)*log(1.0-a)/6.0;
    value -= ((1.0-b)*(1.0-b)*(1.0-b) - (1.0-a)*(1.0-a)*(1.0-a))/18.0;
    if(fabs(1.0-b)>1e-8) value -= 0.5*(1.0-b)*(1.0-b)*log(1.0-b);
    if(fabs(1.0-a)>1e-8) value += 0.5*(1.0-a)*(1.0-a)*log(1.0-a);
    value += 0.25*((1.0-b)*(1.0-b) - (1.0-a)*(1.0-a));
  }
  return value;
}

double* 
fill_rhs(int n, int nr){
  int i;
  double* rhs;
  rhs = (double*) malloc(n*sizeof(double));
  for(i=0; i<n; i++)
    rhs[i] = get_rhsentry(n,i,nr);
 
  return rhs;
}
