
#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <assert.h>
#include <gtk/gtk.h>
#include <gtkgl/gtkglarea.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <string.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

#include "operations.h"

#include "bem3d.h"

GtkWidget *mainwindow = NULL;
GtkWidget *select_grid_file = NULL;
GtkWidget *select_cluster_file = NULL;
GtkWidget *select_partition_file = NULL;
GtkWidget *select_function_file = NULL;
GtkWidget *progress_window = NULL;
GtkWidget *status_dialog = NULL;
GtkWidget *popup_menu = NULL;

pbemgrid3d gr = NULL;

double *val = NULL;
int dim = 0;

pclustertree ct = NULL;
pcluster *cstack = NULL;
int cptr = 0;
int cmax = 0;
int cson = 0;
int csonvisible = 0;

int *gflag = NULL;

static gfloat mid[3];
static gfloat rad;

static int lastx = 0;
static int lasty = 0;
static int startx = 0;
static int starty = 0;
static int currentx = 0;
static int currenty = 0;

static gint gllist = 0;

static int drawmode = 2;
static int drawclusters = 0;
static int drawfunction = 0;
static int drawgridlines = 1;

static void
center_transformation(int vertices, const double (*x)[3])
{
  double m[3];
  double r, rx;
  int i;

  m[0] = 0.0;
  m[1] = 0.0;
  m[2] = 0.0;
  for(i=0; i<vertices; i++) {
    m[0] += x[i][0];
    m[1] += x[i][1];
    m[2] += x[i][2];
  }
  m[0] /= vertices;
  m[1] /= vertices;
  m[2] /= vertices;

  r = 0.0;
  for(i=0; i<vertices; i++) {
    rx = sqrt(dsqr(x[i][0] - m[0]) +
	      dsqr(x[i][1] - m[1]) +
	      dsqr(x[i][2] - m[2]));
    if(rx > r)
      r = rx;
  }

  mid[0] = m[0];
  mid[1] = m[1];
  mid[2] = m[2];
  rad = r;
}

static void
drawgl()
{
  GtkWidget *widget;

  widget = lookup_widget(mainwindow, "main_opengl");

  if(gtk_gl_area_make_current(GTK_GL_AREA(widget))) {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glPushMatrix();
    
    glRotated((gfloat) currenty, 1.0, 0.0, 0.0);
    glRotated((gfloat) currentx, 0.0, 1.0, 0.0);

    if(gllist > 0)
      glCallList(gllist);

    glPopMatrix();

    glFlush();

    gtk_gl_area_swapbuffers(GTK_GL_AREA(widget));
  }
}

static void
draw_vertices()
{
  const double (*x)[3] = (const double (*)[3]) gr->x;
  int vertices = gr->vertices;
  int i;

  if(gr != NULL) {
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glScalef(1.0 / rad, 1.0 / rad, 1.0 / rad);
    glTranslatef(-mid[0], -mid[1], -mid[2]);
    
    glColor3f(1.0, 1.0, 1.0);
    
    glBegin(GL_POINTS);
    for(i=0; i<vertices; i++)
      glVertex3f(x[i][0], x[i][1], x[i][2]);
    glEnd();
    
    glPopMatrix();
  }
}

static void
draw_edges()
{
  const double (*x)[3] = (const double (*)[3]) gr->x;
  const int (*e)[2] = (const int (*)[2]) gr->e;
  int edges = gr->edges;
  int i;

  if(gr != NULL) {
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glScalef(1.0 / rad, 1.0 / rad, 1.0 / rad);
    glTranslatef(-mid[0], -mid[1], -mid[2]);
    
    glColor3f(1.0, 1.0, 1.0);
    
    glBegin(GL_LINES);
    for(i=0; i<edges; i++) {
      glVertex3f(x[e[i][0]][0], x[e[i][0]][1], x[e[i][0]][2]);
      glVertex3f(x[e[i][1]][0], x[e[i][1]][1], x[e[i][1]][2]);
    }
    glEnd();
    
    glPopMatrix();
  }
}

static void
draw_triangles()
{
  const double (*x)[3] = (const double (*)[3]) gr->x;
  const int (*t)[3] = (const int (*)[3]) gr->t;
  const double (*n)[3] = (const double (*)[3]) gr->n;
  int triangles = gr->triangles;
  int i;

  if(gr != NULL) {
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glScalef(1.0 / rad, 1.0 / rad, 1.0 / rad);
    glTranslatef(-mid[0], -mid[1], -mid[2]);
    
    glEnable(GL_LIGHTING);
    glColor3f(0.0, 0.0, 1.0);
    glPolygonMode(GL_FRONT, GL_FILL);
    
    glBegin(GL_TRIANGLES);
    for(i=0; i<triangles; i++) {
      glNormal3f(n[i][0], n[i][1], n[i][2]);
      
      glVertex3f(x[t[i][0]][0], x[t[i][0]][1], x[t[i][0]][2]);
      glVertex3f(x[t[i][1]][0], x[t[i][1]][1], x[t[i][1]][2]);
      glVertex3f(x[t[i][2]][0], x[t[i][2]][1], x[t[i][2]][2]);
    }
    glEnd();
    
    if(drawgridlines) {
      glColor3f(0.0, 0.0, 0.0);
      glPolygonMode(GL_FRONT, GL_LINE);
      glLineWidth(0.5);
      glPolygonOffset(-1.0, -1.0);
      glEnable(GL_POLYGON_OFFSET_LINE);
      
      glBegin(GL_TRIANGLES);
      for(i=0; i<triangles; i++) {
	glNormal3f(n[i][0], n[i][1], n[i][2]);
	
	glVertex3f(x[t[i][0]][0], x[t[i][0]][1], x[t[i][0]][2]);
	glVertex3f(x[t[i][1]][0], x[t[i][1]][1], x[t[i][1]][2]);
	glVertex3f(x[t[i][2]][0], x[t[i][2]][1], x[t[i][2]][2]);
      }
      glEnd();
      
      glDisable(GL_POLYGON_OFFSET_LINE);
    }
    
    glDisable(GL_LIGHTING);
    
    glPopMatrix();
  }
}
  
static void
draw_triangle_cluster()
{
  const double (*x)[3] = (const double (*)[3]) gr->x;
  const int (*t)[3] = (const int (*)[3]) gr->t;
  const double (*n)[3] = (const double (*)[3]) gr->n;
  const int *idx = ct->dof2idx;
  int triangles = gr->triangles;
  pcluster tau, tau1;
  int i;

  if(gr != NULL && ct != NULL) {
    for(i=0; i<triangles; i++)
      gflag[i] = 0;

    tau = cstack[cptr];
    for(i=0; i<tau->size; i++)
      gflag[idx[i + tau->start]] = 1;

    if(0 <= cson && cson < tau->sons && csonvisible) {
      tau1 = tau->son[cson];
      for(i=0; i<tau1->size; i++)
	gflag[idx[i + tau1->start]] = 2;
    }

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glScalef(1.0 / rad, 1.0 / rad, 1.0 / rad);
    glTranslatef(-mid[0], -mid[1], -mid[2]);

    glEnable(GL_LIGHTING);
    glPolygonMode(GL_FRONT, GL_FILL);

    glBegin(GL_TRIANGLES);
    for(i=0; i<triangles; i++) {
      glNormal3f(n[i][0], n[i][1], n[i][2]);

      switch(gflag[i]) {
      default:
      case 0:
	glColor3f(0.0, 0.0, 1.0);
	break;
      case 1:
	glColor3f(1.0, 0.0, 0.0);
	break;
      case 2:
	glColor3f(1.0, 0.4, 0.0);
	break;
      }

      glVertex3f(x[t[i][0]][0], x[t[i][0]][1], x[t[i][0]][2]);
      glVertex3f(x[t[i][1]][0], x[t[i][1]][1], x[t[i][1]][2]);
      glVertex3f(x[t[i][2]][0], x[t[i][2]][1], x[t[i][2]][2]);
    }
    glEnd();

    if(drawgridlines) {
      glColor3f(0.0, 0.0, 0.0);
      glPolygonMode(GL_FRONT, GL_LINE);
      glLineWidth(0.5);
      glPolygonOffset(-1.0, -1.0);
      glEnable(GL_POLYGON_OFFSET_LINE);

      glBegin(GL_TRIANGLES);
      for(i=0; i<triangles; i++) {
	glNormal3f(n[i][0], n[i][1], n[i][2]);
      
	glVertex3f(x[t[i][0]][0], x[t[i][0]][1], x[t[i][0]][2]);
	glVertex3f(x[t[i][1]][0], x[t[i][1]][1], x[t[i][1]][2]);
	glVertex3f(x[t[i][2]][0], x[t[i][2]][1], x[t[i][2]][2]);
      }
      glEnd();

      glDisable(GL_POLYGON_OFFSET_LINE);
    }

    glDisable(GL_LIGHTING);

    glPopMatrix();
  }
}

static void
draw_edge_cluster()
{
  const double (*x)[3] = (const double (*)[3]) gr->x;
  const int (*t)[3] = (const int (*)[3]) gr->t;
  const int (*e)[2] = (const int (*)[2]) gr->e;
  const double (*n)[3] = (const double (*)[3]) gr->n;
  const int *idx = ct->dof2idx;
  int vertices = gr->vertices;
  int triangles = gr->triangles;
  pcluster tau, tau1;
  int i;

  if(gr != NULL && ct != NULL) {
    for(i=0; i<vertices; i++)
      gflag[i] = 0;

    tau = cstack[cptr];
    for(i=0; i<tau->size; i++) {
      gflag[e[idx[i+tau->start]][0]] = 1;
      gflag[e[idx[i+tau->start]][1]] = 1;
    }

    if(0 <= cson && cson < tau->sons && csonvisible) {
      tau1 = tau->son[cson];
      for(i=0; i<tau1->size; i++) {
	gflag[e[idx[i+tau1->start]][0]] = 2;
	gflag[e[idx[i+tau1->start]][1]] = 2;
      }
    }

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glScalef(1.0 / rad, 1.0 / rad, 1.0 / rad);
    glTranslatef(-mid[0], -mid[1], -mid[2]);

    glEnable(GL_LIGHTING);
    glPolygonMode(GL_FRONT, GL_FILL);

    glBegin(GL_TRIANGLES);
    for(i=0; i<triangles; i++) {
      glNormal3f(n[i][0], n[i][1], n[i][2]);

      switch(gflag[t[i][0]]) {
      default:
      case 0:
	glColor3f(0.0, 0.0, 1.0);
	break;
      case 1:
	glColor3f(1.0, 0.0, 0.0);
	break;
      case 2:
	glColor3f(1.0, 0.4, 0.0);
	break;
      }
      glVertex3f(x[t[i][0]][0], x[t[i][0]][1], x[t[i][0]][2]);

      switch(gflag[t[i][1]]) {
      default:
      case 0:
	glColor3f(0.0, 0.0, 1.0);
	break;
      case 1:
	glColor3f(1.0, 0.0, 0.0);
	break;
      case 2:
	glColor3f(1.0, 0.4, 0.0);
	break;
      }
      glVertex3f(x[t[i][1]][0], x[t[i][1]][1], x[t[i][1]][2]);

      switch(gflag[t[i][2]]) {
      default:
      case 0:
	glColor3f(0.0, 0.0, 1.0);
	break;
      case 1:
	glColor3f(1.0, 0.0, 0.0);
	break;
      case 2:
	glColor3f(1.0, 0.4, 0.0);
	break;
      }
      glVertex3f(x[t[i][2]][0], x[t[i][2]][1], x[t[i][2]][2]);
    }
    glEnd();

    if(drawgridlines) {
      glColor3f(0.0, 0.0, 0.0);
      glPolygonMode(GL_FRONT, GL_LINE);
      glLineWidth(0.5);
      glPolygonOffset(-1.0, -1.0);
      glEnable(GL_POLYGON_OFFSET_LINE);

      glBegin(GL_TRIANGLES);
      for(i=0; i<triangles; i++) {
	glNormal3f(n[i][0], n[i][1], n[i][2]);
      
	glVertex3f(x[t[i][0]][0], x[t[i][0]][1], x[t[i][0]][2]);
	glVertex3f(x[t[i][1]][0], x[t[i][1]][1], x[t[i][1]][2]);
	glVertex3f(x[t[i][2]][0], x[t[i][2]][1], x[t[i][2]][2]);
      }
      glEnd();

      glDisable(GL_POLYGON_OFFSET_LINE);
    }

    glDisable(GL_LIGHTING);

    glPopMatrix();
  }
}

static void
draw_vertex_cluster()
{
  const double (*x)[3] = (const double (*)[3]) gr->x;
  const int (*t)[3] = (const int (*)[3]) gr->t;
  const double (*n)[3] = (const double (*)[3]) gr->n;
  const int *idx = ct->dof2idx;
  int vertices = gr->vertices;
  int triangles = gr->triangles;
  pcluster tau, tau1;
  int i;

  if(gr != NULL && ct != NULL) {
    for(i=0; i<vertices; i++)
      gflag[i] = 0;

    tau = cstack[cptr];
    for(i=0; i<tau->size; i++)
      gflag[idx[i + tau->start]] = 1;

    if(0 <= cson && cson < tau->sons && csonvisible) {
      tau1 = tau->son[cson];
      for(i=0; i<tau1->size; i++)
	gflag[idx[i + tau1->start]] = 2;
    }

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glScalef(1.0 / rad, 1.0 / rad, 1.0 / rad);
    glTranslatef(-mid[0], -mid[1], -mid[2]);

    glEnable(GL_LIGHTING);
    glPolygonMode(GL_FRONT, GL_FILL);

    glBegin(GL_TRIANGLES);
    for(i=0; i<triangles; i++) {
      glNormal3f(n[i][0], n[i][1], n[i][2]);

      switch(gflag[t[i][0]]) {
      default:
      case 0:
	glColor3f(0.0, 0.0, 1.0);
	break;
      case 1:
	glColor3f(1.0, 0.0, 0.0);
	break;
      case 2:
	glColor3f(1.0, 0.4, 0.0);
	break;
      }
      glVertex3f(x[t[i][0]][0], x[t[i][0]][1], x[t[i][0]][2]);

      switch(gflag[t[i][1]]) {
      default:
      case 0:
	glColor3f(0.0, 0.0, 1.0);
	break;
      case 1:
	glColor3f(1.0, 0.0, 0.0);
	break;
      case 2:
	glColor3f(1.0, 0.4, 0.0);
	break;
      }
      glVertex3f(x[t[i][1]][0], x[t[i][1]][1], x[t[i][1]][2]);

      switch(gflag[t[i][2]]) {
      default:
      case 0:
	glColor3f(0.0, 0.0, 1.0);
	break;
      case 1:
	glColor3f(1.0, 0.0, 0.0);
	break;
      case 2:
	glColor3f(1.0, 0.4, 0.0);
	break;
      }
      glVertex3f(x[t[i][2]][0], x[t[i][2]][1], x[t[i][2]][2]);
    }
    glEnd();

    if(drawgridlines) {
      glColor3f(0.0, 0.0, 0.0);
      glPolygonMode(GL_FRONT, GL_LINE);
      glLineWidth(0.5);
      glPolygonOffset(-1.0, -1.0);
      glEnable(GL_POLYGON_OFFSET_LINE);

      glBegin(GL_TRIANGLES);
      for(i=0; i<triangles; i++) {
	glNormal3f(n[i][0], n[i][1], n[i][2]);
      
	glVertex3f(x[t[i][0]][0], x[t[i][0]][1], x[t[i][0]][2]);
	glVertex3f(x[t[i][1]][0], x[t[i][1]][1], x[t[i][1]][2]);
	glVertex3f(x[t[i][2]][0], x[t[i][2]][1], x[t[i][2]][2]);
      }
      glEnd();

      glDisable(GL_POLYGON_OFFSET_LINE);
    }

    glDisable(GL_LIGHTING);

    glPopMatrix();
  }
}

static void
draw_triangle_function()
{
  const double (*x)[3] = (const double (*)[3]) gr->x;
  const int (*t)[3] = (const int (*)[3]) gr->t;
  const double (*n)[3] = (const double (*)[3]) gr->n;
  int triangles = gr->triangles;
  int i;

  if(gr != NULL && val != NULL ) {
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glScalef(1.0 / rad, 1.0 / rad, 1.0 / rad);
    glTranslatef(-mid[0], -mid[1], -mid[2]);

    glEnable(GL_LIGHTING);
    glPolygonMode(GL_FRONT, GL_FILL);

    glBegin(GL_TRIANGLES);
    for(i=0; i<triangles; i++) {
      glNormal3f(n[i][0], n[i][1], n[i][2]);

      glColor3f(val[i], 0.0, 1.0-val[i]);

      glVertex3f(x[t[i][0]][0], x[t[i][0]][1], x[t[i][0]][2]);
      glVertex3f(x[t[i][1]][0], x[t[i][1]][1], x[t[i][1]][2]);
      glVertex3f(x[t[i][2]][0], x[t[i][2]][1], x[t[i][2]][2]);
    }
    glEnd();

    if(drawgridlines) {
      glColor3f(0.0, 0.0, 0.0);
      glPolygonMode(GL_FRONT, GL_LINE);
      glLineWidth(0.5);
      glPolygonOffset(-1.0, -1.0);
      glEnable(GL_POLYGON_OFFSET_LINE);

      glBegin(GL_TRIANGLES);
      for(i=0; i<triangles; i++) {
	glNormal3f(n[i][0], n[i][1], n[i][2]);
      
	glVertex3f(x[t[i][0]][0], x[t[i][0]][1], x[t[i][0]][2]);
	glVertex3f(x[t[i][1]][0], x[t[i][1]][1], x[t[i][1]][2]);
	glVertex3f(x[t[i][2]][0], x[t[i][2]][1], x[t[i][2]][2]);
      }
      glEnd();

      glDisable(GL_POLYGON_OFFSET_LINE);
    }

    glDisable(GL_LIGHTING);

    glPopMatrix();
  }
}

static void
draw_vertex_function()
{
  const double (*x)[3] = (const double (*)[3]) gr->x;
  const int (*t)[3] = (const int (*)[3]) gr->t;
  const double (*n)[3] = (const double (*)[3]) gr->n;
  int triangles = gr->triangles;
  int i, v;

  if(gr != NULL && val != NULL) {
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glScalef(1.0 / rad, 1.0 / rad, 1.0 / rad);
    glTranslatef(-mid[0], -mid[1], -mid[2]);

    glEnable(GL_LIGHTING);
    glPolygonMode(GL_FRONT, GL_FILL);

    glBegin(GL_TRIANGLES);
    for(i=0; i<triangles; i++) {
      glNormal3f(n[i][0], n[i][1], n[i][2]);

      v = t[i][0];
      glColor3f(val[v], 0.0, 1.0-val[v]);
      glVertex3f(x[v][0], x[v][1], x[v][2]);

      v = t[i][1];
      glColor3f(val[v], 0.0, 1.0-val[v]);
      glVertex3f(x[v][0], x[v][1], x[v][2]);
      
      v = t[i][2];
      glColor3f(val[v], 0.0, 1.0-val[v]);
      glVertex3f(x[v][0], x[v][1], x[v][2]);
    }
    glEnd();

    if(drawgridlines) {
      glColor3f(0.0, 0.0, 0.0);
      glPolygonMode(GL_FRONT, GL_LINE);
      glLineWidth(0.5);
      glPolygonOffset(-1.0, -1.0);
      glEnable(GL_POLYGON_OFFSET_LINE);

      glBegin(GL_TRIANGLES);
      for(i=0; i<triangles; i++) {
	glNormal3f(n[i][0], n[i][1], n[i][2]);
      
	glVertex3f(x[t[i][0]][0], x[t[i][0]][1], x[t[i][0]][2]);
	glVertex3f(x[t[i][1]][0], x[t[i][1]][1], x[t[i][1]][2]);
	glVertex3f(x[t[i][2]][0], x[t[i][2]][1], x[t[i][2]][2]);
      }
      glEnd();

      glDisable(GL_POLYGON_OFFSET_LINE);
    }

    glDisable(GL_LIGHTING);

    glPopMatrix();
  }
}

static void
draw_grid()
{
  GtkWidget *widget;

  widget = lookup_widget(mainwindow, "main_opengl");

  if(gtk_gl_area_make_current(GTK_GL_AREA(widget))) {
    if(gllist > 0)
      glDeleteLists(gllist, 1);
    
    gllist = glGenLists(1);
    
    glNewList(gllist, GL_COMPILE);
    
    if(gr)
      switch(drawmode) {
      case 0:
	draw_vertices();
	break;
      case 1:
	draw_edges();
	break;
      default:
      case 2:
	draw_triangles();
	break;
      case 3:
	switch(drawclusters) {
	case 0:
	  draw_triangle_cluster();
	  break;
	case 1:
	  draw_edge_cluster();
	  break;
	case 2:
	  draw_vertex_cluster();
	  break;
	}
	break;
      case 4:
	switch(drawfunction) {
	case 0:
	  draw_triangle_function();
	  break;
	case 2:
	  draw_vertex_function();
	  break;
	}
	break;
      }

    glEndList();
  }
}

void
on_read_grid                           (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gtk_widget_show(select_grid_file);
}


void
on_display_switched                    (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkLabel *label;

  label = GTK_LABEL(GTK_BIN(menuitem)->child);

  if(strcmp(label->label, "Vertices") == 0)
    drawmode = 0;
  else
    if(strcmp(label->label, "Edges") == 0)
      drawmode = 1;
    else
      if(strcmp(label->label, "Triangles") == 0)
	drawmode = 2;
      else
	if(strcmp(label->label, "Clusters") == 0)
	  drawmode = 3;
	else
	  drawmode = 4;

  draw_grid();
  drawgl();
}


void
on_read_cluster                        (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gtk_widget_show(select_cluster_file);
}

void
on_bounding_boxes                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}

void
on_cluster_father                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if(ct && cptr > 0) {
    cptr--;

    draw_grid();
    drawgl();
  }
}

void
on_cluster_son                         (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if(ct && 0 <= cson && cson < cstack[cptr]->sons && cptr+1 < cmax) {
    cstack[cptr+1] = cstack[cptr]->son[cson];
    cptr++;

    draw_grid();
    drawgl();
  }
}

static gint
reset_sonvisible(gpointer data)
{
  (void) data;

  if(csonvisible > 0) {
    csonvisible--;

    if(csonvisible == 0) {
      draw_grid();
      drawgl();

      return FALSE;
    }
  }

  return TRUE;
}

void
on_cluster_next                        (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if(ct && cstack[cptr]->sons > 0) {
    cson = (cson + 1) % cstack[cptr]->sons;

    if(csonvisible == 0)
      gtk_timeout_add(100, reset_sonvisible, NULL);

    csonvisible = 20;

    draw_grid();
    drawgl();
  }
}

void
on_cluster_prev                        (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if(ct && cstack[cptr]->sons > 0) {
    cson = (cson + cstack[cptr]->sons - 1) % cstack[cptr]->sons;

    if(csonvisible == 0)
      gtk_timeout_add(100, reset_sonvisible, NULL);

    csonvisible = 20;

    draw_grid();
    drawgl();
  }
}

void
on_read_partition                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gtk_widget_show(select_partition_file);
}

void
on_block_row                           (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}

GtkWidget*
create_opengl (gchar *widget_name, gchar *string1, gchar *string2,
                gint int1, gint int2)
{
  GtkWidget *opengl_widget;
  int attrlist[] = { GDK_GL_RGBA, GDK_GL_DOUBLEBUFFER, GDK_GL_NONE };

  opengl_widget = gtk_gl_area_new(attrlist);
  assert(opengl_widget != NULL);

  return opengl_widget;
}

static void
reset_clustertree()
{
  if(ct)
    del_clustertree(ct);
  ct = NULL;

  if(cstack != NULL)
    free(cstack);
  cstack = NULL;

  cmax = 0;
  cptr = 0;
  cson = 0;
}

static void
reset_grid()
{
  GtkWidget *widget;

  if(gr)
    del_bemgrid3d(gr);
  gr = NULL;
  
  if(gflag)
    free(gflag);
  gflag = NULL;
    
  reset_clustertree();

  if(drawmode > 2) {
    drawmode = 2;

    widget = lookup_widget(mainwindow, "view_triangles");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(widget), TRUE);
  }
}

void
on_grid_file_selected                  (GtkButton       *button,
                                        gpointer         user_data)
{
  pbemgrid3d grnew;
  gchar *filename;

  gtk_widget_hide(select_grid_file);

  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(select_grid_file));

  grnew = read_bemgrid3d_graphical(filename);

  if(grnew != NULL) {
    reset_grid();

    gr = grnew;
    gflag = (int *) malloc((size_t) sizeof(int) * imax(gr->triangles,
						       imax(gr->edges, gr->vertices)));
  }

  while(gtk_events_pending())
    gtk_main_iteration();

  prepare_bemgrid3d(gr);
  center_transformation(gr->vertices, (const double (*)[3]) gr->x);

  if(!closed_bemgrid3d(gr))
    g_warning("Grid is not closed.\n");
  else
    if(!oriented_bemgrid3d(gr))
      g_warning("Grid is closed, but not oriented.\n");

  draw_grid();
  drawgl();
}

void
on_cluster_file_selected               (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *widget;
  pclustertree ctnew;
  gchar *filename;

  gtk_widget_hide(select_cluster_file);

  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(select_cluster_file));

  ctnew = read_clustertree(filename);

  if(ctnew != NULL) {
    reset_clustertree();

    ct = ctnew;

    cmax = depth_cluster(ct->root);
    cstack = (pcluster *) malloc((size_t) sizeof(pcluster) * cmax);
    cstack[0] = ct->root;
    cptr = 0;

    if(ct->nidx == gr->triangles) {
      drawmode = 3;
      drawclusters = 0;

      widget = lookup_widget(mainwindow, "view_clusters");
      gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(widget), TRUE);
    }
    else if(ct->nidx == gr->edges) {
      drawmode = 3;
      drawclusters = 1;

      widget = lookup_widget(mainwindow, "view_clusters");
      gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(widget), TRUE);
    }
    else if(ct->nidx == gr->vertices) {
      drawmode = 3;
      drawclusters = 2;

      widget = lookup_widget(mainwindow, "view_clusters");
      gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(widget), TRUE);
    }
    else {
      drawmode = 2;

      widget = lookup_widget(mainwindow, "view_triangles");
      gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(widget), TRUE);
    }
  }

  while(gtk_events_pending())
    gtk_main_iteration();

  draw_grid();
  drawgl();
}

void
on_partition_file_selected             (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_hide(select_partition_file);
}

gboolean
on_opengl_configure                    (GtkWidget       *widget,
                                        GdkEventConfigure *event,
                                        gpointer         user_data)
{
  GLsizei w = (GLsizei) widget->allocation.width;
  GLsizei h = (GLsizei) widget->allocation.height;
  GLfloat color[4], pos[4];

  if(gtk_gl_area_make_current(GTK_GL_AREA(widget))) {
    glViewport(0, 0, w, h);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    if(w > h)
      glFrustum(-1.2*w/h, 1.2*w/h, -1.2, 1.2, 10.0, 1000.0);
    else
      glFrustum(-1.2, 1.2, -1.2*h/w, 1.2*h/w, 10.0, 1000.0);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(0.0, 0.0, 12.0,
	      0.0, 0.0, 0.0,
	      0.0, 1.0, 0.0);

    color[0] = color[1] = color[2] = 0.3; color[3] = 1.0;
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, color);
    color[0] = color[2] = color[3] = 1.0; color[1] = 0.0;
    glMaterialfv(GL_BACK, GL_DIFFUSE, color);
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 100.0);
    glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);

    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
    pos[0] = -1.0; pos[1] = 1.0; pos[2] = 1.0; pos[3] = 0.0;
    glLightfv(GL_LIGHT0, GL_POSITION, pos);
    color[0] = 0.1; color[1] = 0.1; color[2] = 0.1; color[3] = 1.0;
    glLightfv(GL_LIGHT0, GL_AMBIENT, color);
    glEnable(GL_LIGHT0);
  }
  else
    g_error("Couldn't configure OpenGL area.");

  return TRUE;
}

void
on_opengl_realize                      (GtkWidget       *widget,
                                        gpointer         user_data)
{
  if(gtk_gl_area_make_current(GTK_GL_AREA(widget))) {
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glShadeModel(GL_SMOOTH);
    glEnable(GL_DEPTH_TEST);
    glPolygonMode(GL_FRONT, GL_FILL);
    glCullFace(GL_BACK);
    glEnable(GL_CULL_FACE);
    glEnable(GL_RESCALE_NORMAL);
    glDisable(GL_BLEND);
  }
  else
    g_error("Couldn't realize OpenGL area.");
}

gboolean
on_opengl_expose                       (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  drawgl();

  return TRUE;
}

gboolean
on_opengl_press                        (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
  gint x, y;
  guint state;

  /* Apparently necessary, taken from the viewlw.c demo program */
  gdk_window_get_pointer(event->window, &x, &y, &state);

  if(state & GDK_BUTTON1_MASK) {
    startx = x;
    starty = y;
    lastx = currentx;
    lasty = currenty;

    return TRUE;
  }

  if(state & GDK_BUTTON3_MASK) {
    gtk_menu_popup(GTK_MENU(popup_menu),
		   NULL, NULL, NULL, NULL, GDK_BUTTON3_MASK, 0);
  }

  return FALSE;
}

gboolean
on_opengl_motion                       (GtkWidget       *widget,
                                        GdkEventMotion  *event,
                                        gpointer         user_data)
{
  gint x, y;
  guint state;

  /* Apparently necessary, taken from the viewlw.c demo program */
  gdk_window_get_pointer(event->window, &x, &y, &state);

  if(state & GDK_BUTTON1_MASK) {
    currentx = lastx + (x - startx);
    currenty = lasty + (y - starty);

    drawgl();

    return TRUE;
  }

  return FALSE;
}

void
on_grid_lines                          (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  drawgridlines = GTK_CHECK_MENU_ITEM(menuitem)->active;

  draw_grid();
  drawgl();
}

void
on_file_status                         (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *widget;
  char buf[80];

  widget = lookup_widget(status_dialog, "status_vertices");
  if(gr == NULL) 
    gtk_label_set_text(GTK_LABEL(widget), "n/a");
  else {
    (void) sprintf(buf, "%d", gr->vertices);
    gtk_label_set_text(GTK_LABEL(widget), buf);
  }

  widget = lookup_widget(status_dialog, "status_edges");
  if(gr == NULL) 
    gtk_label_set_text(GTK_LABEL(widget), "n/a");
  else {
    (void) sprintf(buf, "%d", gr->edges);
    gtk_label_set_text(GTK_LABEL(widget), buf);
  }

  widget = lookup_widget(status_dialog, "status_triangles");
  if(gr == NULL) 
    gtk_label_set_text(GTK_LABEL(widget), "n/a");
  else {
    (void) sprintf(buf, "%d", gr->triangles);
    gtk_label_set_text(GTK_LABEL(widget), buf);
  }

  widget = lookup_widget(status_dialog, "status_clusters");
  if(ct == NULL) 
    gtk_label_set_text(GTK_LABEL(widget), "n/a");
  else {
    (void) sprintf(buf, "%d", count_cluster(ct->root));
    gtk_label_set_text(GTK_LABEL(widget), buf);
  }

  widget = lookup_widget(status_dialog, "status_dofs");
  gtk_label_set_text(GTK_LABEL(widget), (drawclusters == 0 ?
					 "Triangles" :
					 (drawclusters == 1 ?
					  "Edges" :
					  (drawclusters == 2 ?
					   "Vertices" :
					   "?"))));

  widget = lookup_widget(status_dialog, "status_displaymode");
  gtk_label_set_text(GTK_LABEL(widget), (drawmode == 0 ?
					 "Vertices" :
					 (drawmode == 1 ?
					  "Edges" :
					  (drawmode == 2 ?
					   "Triangles" :
					   (drawmode == 3 ?
					    "Clusters" :
					    "?")))));

  gtk_widget_show(status_dialog);
}

void
on_read_function                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gtk_widget_show(select_function_file);
}

void
on_function_file_selected              (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *widget;
  gchar *filename;
  double minval, maxval;
  int dimnew;
  int i;

  gtk_widget_hide(select_function_file);

  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(select_function_file));

  if(val != NULL) {
    free(val);
    val = NULL;
  }

  read_function_graphical(filename, &dimnew, &val);
  dim = dimnew;

  minval = val[0];
  maxval = val[0];
  for(i=1; i<dimnew; i++) {
    minval = dmin(minval, val[i]);
    maxval = dmax(maxval, val[i]);
  }

  if(maxval - minval < 1e-10)
    for(i=0; i<dimnew; i++)
      val[i] = 0.0;
  else
    for(i=0; i<dimnew; i++)
      val[i] = (val[i] - minval) / (maxval - minval);

  if(dimnew == gr->triangles) {
    drawmode = 4;
    drawfunction = 0;

    widget = lookup_widget(mainwindow, "view_function");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(widget), TRUE);
  }
  else if(dimnew == gr->vertices) {
    drawmode = 4;
    drawfunction = 2;

    widget = lookup_widget(mainwindow, "view_function");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(widget), TRUE);
  }
  else {
    drawmode = 2;
    
    widget = lookup_widget(mainwindow, "view_triangles");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(widget), TRUE);
  }

  while(gtk_events_pending())
    gtk_main_iteration();

  draw_grid();
  drawgl();
}

