
#include <assert.h>
#include <ctype.h>
#include <gtk/gtk.h>
#include <gtkgl/gtkglarea.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <string.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "operations.h"

#include "bem3d.h"

pbemgrid3d
read_bemgrid3d_graphical(const char *filename)
{
  pbemgrid3d gr;
  GtkWidget *widget;
  FILE *in;
  char buf[80];
  double (*x)[3];
  int (*e)[2];
  int (*t)[3];
  int (*s)[3];
  int vertices, edges, triangles;
  int new_percent, old_percent;
  int i, line;

  gtk_widget_show(progress_window);

  in = fopen(filename, "r");
  assert(in != NULL);

  (void) fgets(buf, 80, in);
  if(sscanf(buf, "%d %d %d", &vertices, &edges, &triangles) != 3) {
    (void) fclose(in);
    return read2_bemgrid3d_graphical(filename);
  }

  gr = new_bemgrid3d(vertices, edges, triangles);

  x = gr->x;
  e = gr->e;
  t = gr->t;
  s = gr->s;

  line = 1;
 
  widget = lookup_widget(progress_window, "working_label");
  gtk_label_set_text(GTK_LABEL(widget), "Reading Vertices");

  old_percent = 0;
  widget = lookup_widget(progress_window, "working_progressbar");
  gtk_progress_set_percentage(GTK_PROGRESS(widget), 0.01 * old_percent);
  while(gtk_events_pending())
    gtk_main_iteration();

  for(i=0; i<vertices; i++) {
    x[i][0] = x[i][1] = x[i][2] = 0.0;
    (void) fgets(buf, 80, in);
    line++;
    if(sscanf(buf, "%lf %lf %lf", &x[i][0], &x[i][1], &x[i][2]) != 3)
      g_error("Line %d of file \"%s\" does not contain a vertex.\n",
	      line, filename);

    new_percent = (100 * i) / vertices;
    if(new_percent != old_percent) {
      old_percent = new_percent;
      widget = lookup_widget(progress_window, "working_progressbar");
      gtk_progress_set_percentage(GTK_PROGRESS(widget), 0.01 * old_percent);
      while(gtk_events_pending())
	gtk_main_iteration();
    }
  }

  widget = lookup_widget(progress_window, "working_label");
  gtk_label_set_text(GTK_LABEL(widget), "Reading Edges");

  old_percent = 0;
  widget = lookup_widget(progress_window, "working_progressbar");
  gtk_progress_set_percentage(GTK_PROGRESS(widget), 0.01 * old_percent);
  while(gtk_events_pending())
    gtk_main_iteration();

  for(i=0; i<edges; i++) {
    e[i][0] = e[i][1] = 0;
    (void) fgets(buf, 80, in);
    line++;
    if(sscanf(buf, "%d %d", &e[i][0], &e[i][1]) != 2)
      g_error("Line %d of file \"%s\" does not contain an edge.\n",
	      line, filename);

    new_percent = (100 * i) / edges;
    if(new_percent != old_percent) {
      old_percent = new_percent;
      widget = lookup_widget(progress_window, "working_progressbar");
      gtk_progress_set_percentage(GTK_PROGRESS(widget), 0.01 * old_percent);
      while(gtk_events_pending())
	gtk_main_iteration();
    }
  }

  widget = lookup_widget(progress_window, "working_label");
  gtk_label_set_text(GTK_LABEL(widget), "Reading Triangles");

  old_percent = 0;
  widget = lookup_widget(progress_window, "working_progressbar");
  gtk_progress_set_percentage(GTK_PROGRESS(widget), 0.01 * old_percent);
  while(gtk_events_pending())
    gtk_main_iteration();

  for(i=0; i<triangles; i++) {
    t[i][0] = t[i][1] = t[i][2] = 0;
    (void) fgets(buf, 80, in);
    line++;
    if(sscanf(buf, "%d %d %d %d %d %d",
	      &t[i][0], &t[i][1], &t[i][2],
	      &s[i][0], &s[i][1], &s[i][2]) != 6)
      g_error("Line %d of file \"%s\" does not contain a triangle.\n",
	      line, filename);

    new_percent = (100 * i) / triangles;
    if(new_percent != old_percent) {
      old_percent = new_percent;
      widget = lookup_widget(progress_window, "working_progressbar");
      gtk_progress_set_percentage(GTK_PROGRESS(widget), 0.01 * old_percent);
      while(gtk_events_pending())
	gtk_main_iteration();
    }
  }

  (void) fclose(in);

  gtk_widget_hide(progress_window);

  return gr;
}

static void
fetch_line(char *buf, int bufsize, int *line, FILE *in)
{
  (void) fgets(buf, bufsize, in);
  (*line)++;

  while(*buf == '#' || *buf == '\n') {
    (void) fgets(buf, bufsize, in);
    (*line)++;
  }
}

pbemgrid3d
read2_bemgrid3d_graphical(const char *filename)
{
  pbemgrid3d gr;
  GtkWidget *widget;
  FILE *in;
  char buf[80];
  double (*x)[3];
  int (*e)[2];
  int (*t)[3];
  int (*s)[3];
  int vertices, edges, triangles;
  double xx, yy, zz;
  int sv, ev;
  int v0, v1, v2, e0, e1, e2;
  int new_percent, old_percent, entries, allentries;
  int i, line;

  gtk_widget_show(progress_window);

  in = fopen(filename, "r");
  assert(in != NULL);
  line = 0;

  fetch_line(buf, 80, &line, in);
  if(sscanf(buf, "nv %d", &vertices) != 1) {
    (void) fprintf(stderr,
		   "Cannot read number of vertices in line %d of file \"%s\".\n",
		   line, filename);
    return NULL;
  }

  fetch_line(buf, 80, &line, in);
  if(sscanf(buf, "ne %d", &edges) != 1) {
    (void) fprintf(stderr,
		   "Cannot read number of edges in line %d of file \"%s\".\n",
		   line, filename);
    return NULL;
  }

  fetch_line(buf, 80, &line, in);
  if(sscanf(buf, "nt %d", &triangles) != 1) {
    (void) fprintf(stderr,
		   "Cannot read number of triangles in line %d of file \"%s\".\n",
		   line, filename);
    return NULL;
  }

  widget = lookup_widget(progress_window, "working_label");
  gtk_label_set_text(GTK_LABEL(widget), "Reading Grid");

  old_percent = 0;
  widget = lookup_widget(progress_window, "working_progressbar");
  gtk_progress_set_percentage(GTK_PROGRESS(widget), 0.01 * old_percent);
  while(gtk_events_pending())
    gtk_main_iteration();

  allentries = vertices + edges + triangles;
  entries = 0;

  gr = new_bemgrid3d(vertices, edges, triangles);

  x = gr->x;
  e = gr->e;
  t = gr->t;
  s = gr->s;

  do {
    fetch_line(buf, 80, &line, in);

    switch(tolower(*buf)) {
    case 'v':
      xx = yy = zz = 0.0;
      if(sscanf(buf, "v%d x%lf y%lf z%lf", &i, &xx, &yy, &zz) != 4)
	(void) fprintf(stderr,
		       "Line %d of file \"%s\" does not contain a vertex.\n",
		       line, filename);
      else if(i < 0 || i >= vertices) 
	(void) fprintf(stderr,
		       "Vertex number in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else {
	x[i][0] = xx;
	x[i][1] = yy;
	x[i][2] = zz;
      }
      entries++;
      break;
    case 'e':
      sv = ev = 0;
      if(sscanf(buf, "e%d v%d v%d", &i, &sv, &ev) != 3)
	(void) fprintf(stderr,
		       "Line %d of file \"%s\" does not contain an edge.\n",
		       line, filename);
      else if(i < 0 || i >= edges)
	(void) fprintf(stderr,
		       "Edge number in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(sv < 0 || sv >= vertices)
	(void) fprintf(stderr,
		       "Start number in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(ev < 0 || ev >= vertices)
	(void) fprintf(stderr,
		       "End number in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else {
	e[i][0] = sv;
	e[i][1] = ev;
      }
      entries++;
      break;
    case 't':
      v0 = v1 = v2 = e0 = e1 = e2 = 0;
      if(sscanf(buf, "t%d v%d v%d v%d e%d e%d e%d", &i, &v0, &v1, &v2, &e0, &e1, &e2) != 7)
	(void) fprintf(stderr,
		       "Line %d of file \"%s\" does not contain a triangle.\n",
		       line, filename);
      else if(i < 0 || i >= triangles)
	(void) fprintf(stderr,
		       "Triangle number in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(v0 < 0 || v0 >= vertices)
	(void) fprintf(stderr,
		       "First vertex in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(v1 < 0 || v1 >= vertices)
	(void) fprintf(stderr,
		       "Second vertex in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(v2 < 0 || v2 >= vertices)
	(void) fprintf(stderr,
		       "Third vertex in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(e0 < 0 || e0 >= edges)
	(void) fprintf(stderr,
		       "First edge in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(e1 < 0 || e1 >= edges)
	(void) fprintf(stderr,
		       "Second edge in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(e2 < 0 || e2 >= edges)
	(void) fprintf(stderr,
		       "Third edge in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else {
	t[i][0] = v0;
	t[i][1] = v1;
	t[i][2] = v2;
	s[i][0] = e0;
	s[i][1] = e1;
	s[i][2] = e2;
      }
      entries++;
      break;
    default:
      (void) fprintf(stderr,
		     "Unknown line %d in file \"%s\".\n",
		     line, filename);
    }

    new_percent = (100 * entries) / allentries;
    if(new_percent != old_percent) {
      old_percent = new_percent;
      widget = lookup_widget(progress_window, "working_progressbar");
      gtk_progress_set_percentage(GTK_PROGRESS(widget), 0.01 * old_percent);
      while(gtk_events_pending())
	gtk_main_iteration();
    }
  }
  while(!feof(in));

  (void) fclose(in);

  gtk_widget_hide(progress_window);

  return gr;
}

void
read_function_graphical(const char *filename, int *dim, double **val)
{
  GtkWidget *widget;
  FILE *in;
  char buf[80];
  int new_percent, old_percent;
  int i, n, line;

  gtk_widget_show(progress_window);

  in = fopen(filename, "r");
  assert(in != NULL);

  (void) fgets(buf, 80, in);
  if(sscanf(buf, "%d", &n) != 1)
    g_error("Cannot process the first line of file \"%s\".\n",
	    filename);

  *dim = n;

  *val = (double *) malloc((size_t) sizeof(double) * n);
  assert(*val != NULL);

  line = 1;
 
  widget = lookup_widget(progress_window, "working_label");
  gtk_label_set_text(GTK_LABEL(widget), "Reading Function Values");

  old_percent = 0;
  widget = lookup_widget(progress_window, "working_progressbar");
  gtk_progress_set_percentage(GTK_PROGRESS(widget), 0.01 * old_percent);
  while(gtk_events_pending())
    gtk_main_iteration();

  for(i=0; i<n; i++) {
    (void) fgets(buf, 80, in);
    line++;

    if(sscanf(buf, "%lf", (*val)+i) != 1)
      g_error("Line %d of file \"%s\" does not contain a value.\n",
	      line, filename);

    new_percent = (100 * i) / n;
    if(new_percent != old_percent) {
      old_percent = new_percent;
      widget = lookup_widget(progress_window, "working_progressbar");
      gtk_progress_set_percentage(GTK_PROGRESS(widget), 0.01 * old_percent);
      while(gtk_events_pending())
	gtk_main_iteration();
    }
  }

  (void) fclose(in);

  gtk_widget_hide(progress_window);
}
